from abc import abstractmethod
import pyproj
import numpy as np
import xarray as xr
import cf_xarray as _
import pandas as pd
import logging

logger = logging.getLogger(__name__)


def __require_obsdim__(f):
    """
    This decorator is for methods of Traj that require a time or obs dimension to work.
    """

    def wrapper(*args, **kwargs):
        if args[0].obsdim is None:
            raise ValueError(f'{f} requires an obs or time dimension')
        return f(*args, **kwargs)

    return wrapper

def detect_tx_dim(ds):
    if 'lon' in ds:
        return ds.lon
    elif 'longitude' in ds:
        return ds.longitude
    elif 'x' in ds:
        return ds.x
    elif 'X' in ds:
        return ds.X
    else:
        raise ValueError("Could not determine x / lon variable")


class Traj:
    ds: xr.Dataset

    __gcrs__: pyproj.CRS

    obsdim: str
    """
    Name of the dimension along which observations are taken. Usually either `obs` or `time`.
    """
    timedim: str

    def __init__(self, ds, obsdim, timedim):
        self.ds = ds
        self.__gcrs__ = pyproj.CRS.from_epsg(4326)
        self.obsdim = obsdim
        self.timedim = timedim

    @property
    def tx(self):
        """
        Trajectory x coordinates (usually longitude).

        .. see-also:

            * `ref:tlon`
            * `ref:ty`
        """
        return detect_tx_dim(self.ds)

    @property
    def ty(self):
        """
        Trajectory y coordinates (usually latitude).

        .. see-also:

            * `ref:tlat`
            * `ref:tx`
        """
        if 'lat' in self.ds:
            return self.ds.lat
        elif 'latitude' in self.ds:
            return self.ds.latitude
        elif 'y' in self.ds:
            return self.ds.y
        elif 'Y' in self.ds:
            return self.ds.Y
        else:
            raise ValueError("Could not determine y / lat variable")

    @property
    def tlon(self):
        """
        Retrieve the trajectories in geographic coordinates (longitudes).
        """
        if self.crs.is_geographic:
            return self.tx
        else:
            if self.crs is None:
                return self.tx
            else:
                x, _ = self.transform(self.__gcrs__, self.tx, self.ty)
                X = self.tx.copy(deep=False,
                                 data=x)  # TODO: remove grid-mapping.
                return X

    @property
    def tlat(self):
        """
        Retrieve the trajectories in geographic coordinates (latitudes).
        """
        if self.crs.is_geographic:
            return self.ty
        else:
            if self.crs is None:
                return self.ty
            else:
                _, y = self.transform(self.__gcrs__, self.tx, self.ty)
                Y = self.ty.copy(deep=False,
                                 data=y)  # TODO: remove grid-mapping.
                return Y

    def transform(self, to_crs, x, y):
        """
        Transform coordinates in this datasets coordinate system to `to_crs` coordinate system.

        Args:

            to_crs: `pyproj.CRS`.

            x, y: Coordinates in `self` CRS.

        Returns:

            xn, yn: Coordinates in `to_crs`.
        """
        t = pyproj.Transformer.from_crs(self.crs, to_crs, always_xy=True)
        return t.transform(x, y)

    def itransform(self, from_crs, x, y):
        """
        Transform coordinates in `from_crs` coordinate system to this datasets coordinate system.

        Args:

            from_crs: `pyproj.CRS`.

            x, y: Coordinates in `from_crs` CRS.

        Returns:

            xn, yn: Coordinates in this datasets CRS.
        """
        t = pyproj.Transformer.from_crs(from_crs, self.crs, always_xy=True)
        return t.transform(x, y)

    @property
    def crs(self) -> pyproj.CRS:
        """
        Retrieve the Proj.4 CRS from the CF-defined grid-mapping in the dataset.
        """
        if len(self.ds.cf.grid_mapping_names) == 0:
            logger.debug(
                f'No grid-mapping specified, checking if coordinates are lon/lat..'
            )
            if self.tx.name == 'lon' or self.tx.name == 'longitude':
                # assume this is in latlon projection
                return self.__gcrs__
            else:
                # un-projected, assume coordinates are in xy-Cartesian coordinates.
                logger.debug(
                    'Assuming trajectories are in Cartesian coordinates.')
                return None

        else:
            gm = self.ds.cf['grid_mapping']
            logger.debug(f'Constructing CRS from grid_mapping: {gm}')
            return pyproj.CRS.from_cf(gm.attrs)

    def set_crs(self, crs):
        """
        Returns a new dataset with the CF-supported grid-mapping / projection set to `crs`.
        """

        # TODO: Ideally this would be handled by cf-xarray or rio-xarray.

        ds = self.ds.copy()

        if crs is None:
            logger.info(
                'Removing CRS information and defining trajactories as Cartesian / unprojected data.'
            )

            if 'grid_mapping' in self.ds.cf:
                gm = self.ds.cf['grid_mapping']
                ds = ds.drop(gm.name)

            for var in self.ds:
                if 'grid_mapping' in self.ds[var].attrs:
                    del ds[var].attrs['grid_mapping']

            if self.tx.name == 'lon' or self.tx.name == 'longitude':
                logger.warning(
                    f'Renaming geographic {self.tx.name, self.ty.name} coordinates to x, y..'
                )
                ds['x'] = ds[self.tx.name].copy().rename('x')
                ds['y'] = ds[self.ty.name].copy().rename('y')
                assert ds['x'].name == 'x'

                ds = ds.drop_vars([self.tx.name, self.ty.name])

        else:
            gm = crs.to_cf()

            # Create grid mapping variable
            v = xr.DataArray(name=gm['grid_mapping_name'])
            v.attrs = gm
            ds[v.name] = v
            ds[self.tx.name].attrs['grid_mapping'] = v.name
            ds[self.ty.name].attrs['grid_mapping'] = v.name

        return ds

    @abstractmethod
    def is_1d(self):
        pass

    @abstractmethod
    def is_2d(self):
        pass

    def assign_cf_attrs(self,
                        creator_name=None,
                        creator_email=None,
                        title=None,
                        summary=None,
                        **kwargs) -> xr.Dataset:
        """
        Return a new dataset with CF-standard and common attributes set.
        """
        ds = self.ds.copy(deep=True)

        ds['trajectory'] = ds['trajectory'].astype(str)
        ds['trajectory'].attrs = {
            'cf_role': 'trajectory_id',
            'long_name': 'trajectory name'
        }

        ds = ds.assign_attrs({
            'Conventions':
            'CF-1.10',
            'featureType':
            'trajectory',
            'geospatial_lat_min':
            np.nanmin(self.tlat),
            'geospatial_lat_max':
            np.nanmax(self.tlat),
            'geospatial_lon_min':
            np.nanmin(self.tlon),
            'geospatial_lon_max':
            np.nanmax(self.tlon),
            'time_coverage_start':
            pd.to_datetime(
                np.nanmin(ds['time'].values[
                    ds['time'].values != np.datetime64('NaT')])).isoformat(),
            'time_coverage_end':
            pd.to_datetime(
                np.nanmax(ds['time'].values[
                    ds['time'].values != np.datetime64('NaT')])).isoformat(),
        })

        if creator_name:
            ds = ds.assign_attrs(creator_name=creator_name)

        if creator_email:
            ds = ds.assign_attrs(creator_email=creator_email)

        if title:
            ds = ds.assign_attrs(title=title)

        if summary:
            ds = ds.assign_attrs(summary=summary)

        if kwargs is not None:
            ds = ds.assign_attrs(**kwargs)

        return ds

    def index_of_last(self):
        """Find index of last valid position along each trajectory"""
        return np.ma.notmasked_edges(np.ma.masked_invalid(self.ds.lon.values),
                                     axis=1)[1][1]

    def speed(self):
        """Returns the speed [m/s] along trajectories"""

        distance = self.distance_to_next()
        timedelta_seconds = self.time_to_next() / np.timedelta64(1, 's')

        return distance / timedelta_seconds

    def distance_to(self, other):
        """
        Distance between trajectories or a single point.
        """

        other = other.broadcast_like(self.ds)
        geod = pyproj.Geod(ellps='WGS84')
        az_fwd, a2, distance = geod.inv(self.ds.traj.tlon, self.ds.traj.tlat,
                                        other.traj.tlon, other.traj.tlat)

        ds = xr.Dataset()
        ds['distance'] = xr.DataArray(distance,
                                      name='distance',
                                      coords=self.ds.traj.tlon.coords,
                                      attrs={'unit': 'm'})

        ds['az_fwd'] = xr.DataArray(az_fwd,
                                    name='forward azimuth',
                                    coords=self.ds.traj.tlon.coords,
                                    attrs={'unit': 'degrees'})

        ds['az_bwd'] = xr.DataArray(a2,
                                    name='back azimuth',
                                    coords=self.ds.traj.tlon.coords,
                                    attrs={'unit': 'degrees'})

        return ds

    @__require_obsdim__
    def distance_to_next(self):
        """Returns distance in m from one position to the next

           Last time is repeated for last position (which has no next position)
        """

        lon = self.ds.lon
        lat = self.ds.lat
        lenobs = self.ds.sizes[self.obsdim]
        lonfrom = lon.isel({self.obsdim: slice(0, lenobs - 1)})
        latfrom = lat.isel({self.obsdim: slice(0, lenobs - 1)})
        lonto = lon.isel({self.obsdim: slice(1, lenobs)})
        latto = lat.isel({self.obsdim: slice(1, lenobs)})
        geod = pyproj.Geod(ellps='WGS84')
        azimuth_forward, a2, distance = geod.inv(lonfrom, latfrom, lonto,
                                                 latto)

        distance = xr.DataArray(distance, coords=lonfrom.coords, dims=lon.dims)
        distance = xr.concat((distance, distance.isel({self.obsdim: -1})),
                             dim=self.obsdim)  # repeating last time step to
        return distance

    @__require_obsdim__
    def azimuth_to_next(self):
        """Returns azimution travel direction in degrees from one position to the next

           Last time is repeated for last position (which has no next position)
        """

        # TODO: method is almost duplicate of "distance_to_next" above
        lon = self.ds.lon
        lat = self.ds.lat
        lenobs = self.ds.dims[self.obsdim]
        lonfrom = lon.isel({self.obsdim: slice(0, lenobs - 1)})
        latfrom = lat.isel({self.obsdim: slice(0, lenobs - 1)})
        lonto = lon.isel({self.obsdim: slice(1, lenobs)})
        latto = lat.isel({self.obsdim: slice(1, lenobs)})
        geod = pyproj.Geod(ellps='WGS84')
        azimuth_forward, a2, distance = geod.inv(lonfrom, latfrom, lonto,
                                                 latto)

        azimuth_forward = xr.DataArray(azimuth_forward,
                                       coords=lonfrom.coords,
                                       dims=lon.dims)
        azimuth_forward = xr.concat(
            (azimuth_forward, azimuth_forward.isel({self.obsdim: -1})),
            dim=self.obsdim)  # repeating last time step to
        return azimuth_forward

    def velocity_components(self):
        """Returns velocity components [m/s] from one position to the next

           Last time is repeated for last position (which has no next position)
        """
        speed = self.speed()
        azimuth = self.azimuth_to_next()

        u = speed * np.cos(azimuth)
        v = speed * np.sin(azimuth)

        return u, v

    def get_area_convex_hull(self):
        """Returns the area of the convex hull spanned by all particles, per timestep"""

        from scipy.spatial import ConvexHull
        from pyproj import Geod

        area = []
        lons = self.ds.lon.where(self.ds.status == 0)
        lats = self.ds.lat.where(self.ds.status == 0)
        for i in range(self.ds.dims['time']):
            lat = lats.isel(time=i)
            lon = lons.isel(time=i)
            fin = np.isfinite(lat + lon)
            if np.sum(fin) <= 3:
                area.append(0)
                continue
            if len(np.unique(lat)) == 1 and len(np.unique(lon)) == 1:
                area.append(0)
                continue
            lat = lat[fin]
            lon = lon[fin]
            aea = pyproj.Proj(
                f'+proj=aea +lat_0={lat.mean().values} +lat_1={lat.min().values} +lat_2={lat.max().values} +lon_0={lon.mean().values} +x_0=0 +y_0=0 +datum=NAD83 +units=m +no_defs'
            )

            x, y = aea(lat, lon, inverse=False)
            fin = np.isfinite(x + y)
            points = np.vstack((y.T, x.T)).T
            hull = ConvexHull(points)
            area.append(hull.volume)  # volume=area for 2D as here

        return np.array(area)

    @abstractmethod
    def gridtime(self, times, timedim = None):
        """Interpolate dataset to a regular time interval or a different grid.

        Args:

            `times`: Target time interval, can be either:
                - an array of times, or
                - a string "freq" specifying a Pandas daterange (e.g. 'h', '6h, 'D'...) suitable for `pd.date_range`.

            `timedime`: Name of new time dimension. The default is to use the same name as previously.

        Returns:

            A new dataset interpolated to the target times. The dataset will be 1D (i.e. gridded) and the time dimension will be named `time`.
        """



"""
Trajan API
"""
import collections
import numpy as np
import xarray as xr
import pandas as pd

from . import trajectory_accessor as _
from . import skill as _
from . import readers as _


def read_csv(f, **kwargs):
    """
    Construct a CF-compliant trajectory dataset from a CSV file. A thin wrapper around: :meth:`from_dataframe`.
    """
    df = pd.read_csv(f)
    return from_dataframe(df, **kwargs)


def from_dataframe(df: pd.DataFrame,
                   lon='lon',
                   lat='lat',
                   time='time',
                   name=None,
                   *,
                   __test_condense__=False):
    """
    Construct a CF-compliant trajectory dataset from a `pd.DataFrame` of positions.

    Args:

        `lon`: name of column containing longitudes.

        `lat`: name of column containing latitudes.

        `time`: name of column containing timestamps (parsable by pandas).

        `name`: name of column to be used for drifter names.

    Returns:

        `ds`: a CF-compliant trajectory `xarray.Dataset`.


    Constructing a dataset from arrays of postions and time:
    --------------------------------------------------------

    .. testcode::

        import pandas as pd
        import trajan as ta

        # Construct some synthetic data
        lon = np.linspace(5, 10, 50)
        lat = np.linspace(60, 70, 50)
        temp = np.linspace(10, 15, 50)
        time = pd.date_range('2023-01-01', '2023-01-14', periods=50)

        # Construct a pandas.DataFrame
        ds = pd.DataFrame({'lon': lon, 'lat': lat, 'time': time, 'temp': temp, 'name': 'My drifter'})
        ds = ta.from_dataframe(ds, name='name')
        print(ds)

    .. testoutput::

        <xarray.Dataset>
        Dimensions:     (trajectory: 1, obs: 50)
        Coordinates:
          * trajectory  (trajectory) <U10 'My drifter'
          * obs         (obs) int64 0 1 2 3 4 5 6 7 8 9 ... 41 42 43 44 45 46 47 48 49
        Data variables:
            lon         (trajectory, obs) float64 5.0 5.102 5.204 ... 9.796 9.898 10.0
            lat         (trajectory, obs) float64 60.0 60.2 60.41 ... 69.59 69.8 70.0
            time        (trajectory, obs) datetime64[ns] 2023-01-01 ... 2023-01-14
            temp        (trajectory, obs) float64 10.0 10.1 10.2 ... 14.8 14.9 15.0
        Attributes:
            Conventions:          CF-1.10
            featureType:          trajectory
            geospatial_lat_min:   60.0
            geospatial_lat_max:   70.0
            geospatial_lon_min:   5.0
            geospatial_lon_max:   10.0
            time_coverage_start:  2023-01-01T00:00:00
            time_coverage_end:    2023-01-14T00:00:00

    Often you might want to add some attributes:

    .. testcode::

        ds = ds.assign_attrs({'Author': 'Albus Dumbledore'})
        print(ds)

    .. testoutput::

        <xarray.Dataset>
        Dimensions:     (trajectory: 1, obs: 50)
        Coordinates:
          * trajectory  (trajectory) <U10 'My drifter'
          * obs         (obs) int64 0 1 2 3 4 5 6 7 8 9 ... 41 42 43 44 45 46 47 48 49
        Data variables:
            lon         (trajectory, obs) float64 5.0 5.102 5.204 ... 9.796 9.898 10.0
            lat         (trajectory, obs) float64 60.0 60.2 60.41 ... 69.59 69.8 70.0
            time        (trajectory, obs) datetime64[ns] 2023-01-01 ... 2023-01-14
            temp        (trajectory, obs) float64 10.0 10.1 10.2 ... 14.8 14.9 15.0
        Attributes:
            Conventions:          CF-1.10
            featureType:          trajectory
            geospatial_lat_min:   60.0
            geospatial_lat_max:   70.0
            geospatial_lon_min:   5.0
            geospatial_lon_max:   10.0
            time_coverage_start:  2023-01-01T00:00:00
            time_coverage_end:    2023-01-14T00:00:00
            Author:               Albus Dumbledore

    """
    df = df.copy()
    df.index.names = ['obs']
    df[time] = pd.to_datetime(df[time], format='mixed')

    df = df.rename(columns={lat: 'lat', lon: 'lon', time: 'time'})

    if name is not None:
        df = df.rename(columns={name: 'trajectory'})
    else:
        df['trajectory'] = 'Drifter 1'

    # Convert to UTC and remove tz-info. Xarray does not support tz-aware
    # datetimes.
    if df['time'].dt.tz is not None:
        df['time'] = df['time'].dt.tz_convert(None)

    # Classify trajectories based on drifter names.
    df = df.set_index(['trajectory', df.index])
    df = df.to_xarray()

    if not __test_condense__:
        df = df.traj.condense_obs()

    df = df.traj.assign_cf_attrs()

    return df


def trajectory_dict_to_dataset(trajectory_dict,
                               variable_attributes=None,
                               global_attributes=None):
    """Create a CF-compatible trajectory file from dictionary of drifter positions

    Trajectory_dict shall have the following structure:
        {'buoy1_name': {
            time0: {'lon': lon0, 'lat': lat0, 'variable1': var1_0, ... 'variableM': varM_0},
            time1: {'lon': lon1, 'lat': lat1, 'variable1': var1_1, ... 'variableM': varM_1},
                ...
            timeN: {'lon': lonN, 'lat': latN, 'variable1': var1_N, ... 'variableM': varM_N}},
        {'buoy2_name': {
            ...
    """

    if variable_attributes is None:
        variable_attributes = dict()

    if global_attributes is None:
        global_attributes = dict()

    drifter_names = list(trajectory_dict)
    num_drifters = len(drifter_names)
    num_times = np.max([len(d) for dn, d in trajectory_dict.items()])
    # Allocate  arrays
    variables = []
    for dn, d in trajectory_dict.items():
        v = list(d[list(d)[0]])
        for var in v:
            if var not in variables:
                variables.extend([var])
    datavars = {}
    for var in variables:
        datavars[var] = np.empty((num_drifters, num_times))
        datavars[var][:] = np.nan
        if var not in variable_attributes:
            if var == 'lat':
                variable_attributes[var] = {
                    'standard_name': 'latitude',
                    'unit': 'degree_north'
                }
            elif var == 'lon':
                variable_attributes[var] = {
                    'standard_name': 'longitude',
                    'unit': 'degree_east'
                }
            else:
                variable_attributes[var] = {}
    time = np.empty((num_drifters, num_times), dtype='datetime64[s]')
    time[:] = np.datetime64('nat')

    # Fill arrays with data from dictionaries
    for drifter_num, (drifter_name,
                      drifter_dict) in enumerate(trajectory_dict.items()):
        # Make sure that positions are sorted by time
        drifter_dict = collections.OrderedDict(sorted(drifter_dict.items()))
        t = slice(0, len(drifter_dict))
        for var in variables:
            datavars[var][drifter_num, t] = np.array(
                [di[var] for d, di in drifter_dict.items()])
        time[drifter_num, t] = np.array(list(drifter_dict),
                                        dtype='datetime64[s]')

    datavars = {
        v: (['trajectory', 'obs'], datavars[v], variable_attributes[v])
        for v in variables
    }
    datavars['time'] = (['trajectory', 'obs'], time, {'standard_name': 'time'})
    datavars['drifter_names'] = (['trajectory'], drifter_names, {
        'cf_role': 'trajectory_id',
        'standard_name': 'platform_id'
    })

    # Remove empty attributes
    global_attributes = {a: v for a, v in global_attributes.items() if v != ''}

    # Create Xarray Dataset adhering to CF conventions h.4.1 for trajectory data
    ds = xr.Dataset(data_vars=datavars,
                    attrs={
                        'Conventions': 'CF-1.10',
                        'featureType': 'trajectory',
                        'geospatial_lat_min': np.nanmin(datavars['lat'][1]),
                        'geospatial_lat_max': np.nanmax(datavars['lat'][1]),
                        'geospatial_lon_min': np.nanmin(datavars['lon'][1]),
                        'geospatial_lon_max': np.nanmax(datavars['lon'][1]),
                        'time_coverage_start':
                        str(np.nanmin(datavars['time'][1])),
                        'time_coverage_end':
                        str(np.nanmax(datavars['time'][1])),
                        **global_attributes
                    })

    return ds



"""
Extending xarray Dataset with functionality specific to trajectory datasets.

Presently supporting Cf convention H.4.1
https://cfconventions.org/Data/cf-conventions/cf-conventions-1.10/cf-conventions.html#_multidimensional_array_representation_of_trajectories.
"""

import xarray as xr
import logging

from .plot import Plot
from .animation import Animation
from .traj import Traj, detect_tx_dim
from .traj1d import Traj1d
from .traj2d import Traj2d

logger = logging.getLogger(__name__)


@xr.register_dataset_accessor("traj")
class TrajAccessor:
    _ds: xr.Dataset
    __plot__: Plot
    __animate__: Animation
    inner: Traj

    def __init__(self, xarray_obj):
        self._ds = xarray_obj
        self.__plot__ = None
        self.__animate__ = None

        if 'traj' in self.ds.dims:
            logger.info(
                'Normalizing dimension name from "traj" to "trajectory".')
            self._ds = self._ds.rename({'traj': 'trajectory'})

        if 'trajectory' not in self.ds.dims:  # Add empty trajectory dimension, if single trajectory
            self._ds = self._ds.expand_dims({'trajectory': 1})
            self._ds['trajectory'].attrs['cf_role'] = 'trajectory_id'


        obsdim = None
        timedim = None

        tx = detect_tx_dim(self._ds)

        if 'obs' in tx.dims:
            obsdim = 'obs'
            timedim = self.__detect_time_dim__(obsdim)

        elif 'time' in tx.dims:
            obsdim = 'time'
            timedim = 'time'

        else:
            for d in tx.dims:
                if not self.ds[d].attrs.get('cf_role', None) == 'trajectory_id' and not 'traj' in d:

                    obsdim = d
                    timedim = self.__detect_time_dim__(obsdim)

                    break

            if obsdim is None:
                logger.warning('No time or obs dimension detected.')

        logger.debug(f"Detected obs-dim: {obsdim}, detected time-dim: {timedim}.")

        if obsdim is None:
            self.inner = Traj1d(self.ds, obsdim, timedim)

        elif len(self._ds[timedim].shape) <= 1:
            logger.debug('Detected structured (1D) trajectory dataset')
            self.inner = Traj1d(self._ds, obsdim, timedim)

        elif len(self._ds[timedim].shape) == 2:
            logger.debug('Detected un-structured (2D) trajectory dataset')
            self.inner = Traj2d(self._ds, obsdim, timedim)

        else:
            raise ValueError(f'Time dimension has shape greater than 2: {self.ds["timedim"].shape}')

    def __detect_time_dim__(self, obsdim):
        logger.debug(f'Detecting time-dimension for "{obsdim}"..')
        for v in self._ds.variables:
            if obsdim in self._ds[v].dims and 'time' in v:
                return v

        raise ValueError("no time dimension detected")

    @property
    def plot(self):
        if self.__plot__ is None:
            logger.debug(f'Setting up new plot object.')
            self.__plot__ = Plot(self.ds)

        return self.__plot__

    @property
    def animate(self):
        if self.__animate__ is None:
            logger.debug(f'Setting up new animation object.')
            self.__animate__ = Animation(self.ds)

        return self.__animate__

    @property
    def ds(self):
        return self._ds

    def __getattr__(self, attr):
        """
        Forward all other method calls and attributes to traj instance.
        """
        return getattr(self.inner, attr)



import numpy as np
import xarray as xr
import numpy as np
import pandas as pd
import logging
from .traj import Traj, __require_obsdim__
from . import skill

logger = logging.getLogger(__name__)


class Traj1d(Traj):
    """
    A structured dataset, where each trajectory is always given at the same times. Typically the output from a model or from a gridded dataset.
    """

    def __init__(self, ds, obsdim, timedim):
        super().__init__(ds, obsdim, timedim)

    def timestep(self):
        """
        Time step between observations in seconds.
        """
        return ((self.ds.time[1] - self.ds.time[0]) /
                np.timedelta64(1, 's')).values

    def is_1d(self):
        return True

    def is_2d(self):
        return False

    def time_to_next(self):
        time_step = self.ds.time[1] - self.ds.time[0]
        return time_step

    def velocity_spectrum(self):

        if self.ds.sizes['trajectory'] > 1:
            raise ValueError(
                'Spectrum can only be calculated for a single trajectory')

        u, v = self.velocity_components()
        u = u.squeeze()
        v = v.squeeze()
        u = u[np.isfinite(u)]
        v = v[np.isfinite(v)]

        timestep_h = (self.ds.time[1] - self.ds.time[0]) / np.timedelta64(
            1, 'h')  # hours since start

        ps = np.abs(np.fft.rfft(np.abs(u + 1j * v)))
        freq = np.fft.rfftfreq(n=u.size, d=timestep_h.values)
        freq[0] = np.nan

        da = xr.DataArray(
            data=ps,
            name='velocity spectrum',
            dims=['period'],
            coords={'period': (['period'], 1 / freq, {
                'units': 'hours'
            })},
            attrs={'units': 'power'})

        return da

    def rotary_spectrum(self):
        ### TODO unfinished method

        from .tools import rotary_spectra
        if self.ds.sizes['trajectory'] > 1:
            raise ValueError(
                'Spectrum can only be calculated for a single trajectory')

        u, v = self.velocity_components()
        u = u.squeeze()
        v = v.squeeze()
        u = u[np.isfinite(u)]
        v = v[np.isfinite(v)]

        puv, quv, cw, ccw, F = rotary_spectra(u, v)
        print(puv)
        import matplotlib.pyplot as plt
        plt.plot(1 / F, cw)
        plt.xlim([0, 30])
        plt.show()

    @__require_obsdim__
    def skill(self, other, method='liu-weissberg', **kwargs):
        """
        Compare the skill score between this trajectory and `other`.

        Args:

            other: Another trajectory dataset.

            method: skill-score method, currently only liu-weissberg. See :mod:`trajan.skill`.

            **kwargs: passed on to the skill-score method.

        Returns:

            skill: The skill-score in the same dimensions as this dataset.

        .. note::

            The datasets must be sampled (or have observations) at approximately the same timesteps. Consider using :meth:`trajan.traj2d.gridtime` to interpolate one of the datasets to the other.


        .. note::

            The datasets must have the same number of trajectories. If you wish to compare a single trajectory to many others, duplicate it along the trajectory dimension to match the trajectory dimension of the other. See further down for an example.


        .. testcode::

            import xarray as xr
            import trajan as _
            import lzma

            b = lzma.open('examples/barents.nc.xz')
            ds = xr.open_dataset(b)

            other = ds.copy()

            ds = ds.traj.gridtime('1H')

            other = other.traj.gridtime(ds.time)
            skill = ds.traj.skill(other)

            print(skill)

        .. testoutput::

            <xarray.DataArray 'Skill-score' (trajectory: 2)>
            array([1., 1.], dtype=float32)
            Coordinates:
              * trajectory  (trajectory) int64 0 1
            Attributes:
                method:   liu-weissberg


        If you need to broadcast a dataset with a single drifter to one with many you can use `xarray.broadcast` or `xarray.Dataset.broadcast_like`:

        .. note::

            If the other dataset has any other dimensions, on any other variables, you need to exclude those when broadcasting.

        .. testcode::

            b0 = ds.isel(trajectory=0) # `b0` now only has a single drifter (no trajectory dimension)

            b0 = b0.broadcast_like(ds)
            skill = b0.traj.skill(ds)

            print(skill)

        .. testoutput::

            <xarray.DataArray 'Skill-score' (trajectory: 2)>
            array([1.        , 0.60894716], dtype=float32)
            Coordinates:
              * trajectory  (trajectory) int64 0 1
            Attributes:
                method:   liu-weissberg

        """

        if self.ds.sizes['trajectory'] != other.sizes['trajectory']:
            raise ValueError(
                f"There must be the same number of trajectories in the two datasets that are compared. This dataset: {self.ds.sizes['trajectory']}, other: {other.sizes['trajectory']}."
            )

        diff = np.max(
            np.abs((self.ds[self.obsdim] -
                    other[other.traj.obsdim]).astype('timedelta64[s]').astype(
                        np.float64)))

        if not np.isclose(diff, 0):
            raise ValueError(
                f"The two datasets must have approximately equal time coordinates, maximum difference: {diff} seconds. Consider using `gridtime` to interpolate one of the datasets."
            )

        s = np.zeros((self.ds.sizes['trajectory']), dtype=np.float32)

        # ds = self.ds.dropna(dim=self.obsdim)
        # other = other.dropna(dim=other.traj.obsdim)

        ds = self.ds.transpose('trajectory', self.obsdim, ...)
        other = other.transpose('trajectory', other.traj.obsdim, ...)

        lon0 = ds.traj.tlon
        lat0 = ds.traj.tlat
        lon1 = other.traj.tlon
        lat1 = other.traj.tlat

        for ti in range(0, len(s)):
            if method == 'liu-weissberg':
                s[ti] = skill.liu_weissberg(lon0.isel(trajectory=ti),
                                            lat0.isel(trajectory=ti),
                                            lon1.isel(trajectory=ti),
                                            lat1.isel(trajectory=ti), **kwargs)
            else:
                raise ValueError(f"Unknown skill-score method: {method}.")

        return xr.DataArray(s,
                            name='Skill-score',
                            coords={'trajectory': self.ds.trajectory},
                            attrs={'method': method})

    @__require_obsdim__
    def gridtime(self, times, timedim=None, round=True):
        if isinstance(times, str):  # Make time series with given interval
            if round is True:
                start_time = np.nanmin(np.asarray(self.ds.time.dt.floor(times)))
                end_time = np.nanmax(np.asarray(self.ds.time.dt.ceil(times)))
            else:
                start_time = np.nanmin(np.asarray(self.ds.time))
                end_time = np.nanmax(np.asarray(self.ds.time))
            times = pd.date_range(start_time,
                                  end_time,
                                  freq=times,
                                  inclusive='both')

        if not isinstance(times, np.ndarray):
            times = times.to_numpy()

        timedim = self.timedim if timedim is None else timedim

        ds = self.ds

        if self.obsdim != timedim:
            ds = ds.rename({self.obsdim: timedim}).set_index({timedim: timedim})

        _, ui = np.unique(ds[timedim], return_index=True)

        if len(ui) != len(self.ds[timedim]):
            logger.warning('non-unique time points, dropping time-duplicates')

        ds = ds.isel({timedim : ui})
        ds = ds.isel({timedim : np.where(~pd.isna(ds[timedim].values))[0]})
        ds = ds.interp({timedim: times})

        if not 'trajectory' in ds.dims:
            ds = ds.expand_dims('trajectory')

        return ds
